@@ -8,10 +8,10 @@ try:
 except ImportError:
     pygal = None
 from dateutil.relativedelta import relativedelta
-from sql import Null, Literal, Column
-from sql.aggregate import Sum, Min, Count
-from sql.conditionals import Case
-from sql.functions import CurrentTimestamp, DateTrunc
+from sql import Null, Literal, Column, With
+from sql.aggregate import Sum, Max, Min, Count
+from sql.conditionals import Coalesce
+from sql.functions import CurrentTimestamp, DateTrunc, Power, Ceil, Log
 
 from trytond.pool import Pool
 from trytond.model import ModelSQL, ModelView, UnionMixin, fields
@@ -19,14 +19,7 @@ from trytond.tools import grouped_slice, reduce_ids
 from trytond.transaction import Transaction
 from trytond.pyson import Eval, If
 from trytond.wizard import Wizard, StateTransition, StateAction
-
-__all__ = ['Context',
-    'Customer', 'CustomerTimeseries',
-    'Product', 'ProductTimeseries',
-    'Category', 'CategoryTimeseries', 'CategoryTree',
-    'Country', 'CountryTimeseries',
-    'Subdivision', 'SubdivisionTimeseries',
-    'Region', 'OpenRegion']
+from trytond.i18n import lazy_gettext
 
 
 def pairwise(iterable):
@@ -37,24 +30,35 @@ def pairwise(iterable):
 
 class Abstract(ModelSQL):
 
-    company = fields.Many2One('company.company', "Company")
-    number = fields.Integer("#", help="Number of sales")
+    company = fields.Many2One(
+        'company.company', lazy_gettext("sale.msg_sale_reporting_company"))
+    number = fields.Integer(lazy_gettext("sale.msg_sale_reporting_number"),
+        help=lazy_gettext("sale.msg_sale_reporting_number_help"))
     revenue = fields.Numeric(
-        "Revenue", digits=(16, Eval('currency_digits', 2)),
+        lazy_gettext("sale.msg_sale_reporting_revenue"),
+        digits=(16, Eval('currency_digits', 2)),
         depends=['currency_digits'])
     revenue_trend = fields.Function(
-        fields.Char("Revenue Trend"), 'get_trend')
+        fields.Char(lazy_gettext("sale.msg_sale_reporting_revenue_trend")),
+        'get_trend')
     time_series = None
 
+    currency = fields.Function(fields.Many2One(
+            'currency.currency',
+            lazy_gettext("sale.msg_sale_reporting_currency")),
+        'get_currency')
     currency_digits = fields.Function(
-        fields.Integer("Currency Digits"), 'get_currency_digits')
+        fields.Integer(
+            lazy_gettext("sale.msg_sale_reporting_currency_digits")),
+        'get_currency_digits')
 
     @classmethod
     def table_query(cls):
-        from_item, tables = cls._joins()
-        return from_item.select(*cls._columns(tables),
-            where=cls._where(tables),
-            group_by=cls._group_by(tables))
+        from_item, tables, withs = cls._joins()
+        return from_item.select(*cls._columns(tables, withs),
+            where=cls._where(tables, withs),
+            group_by=cls._group_by(tables, withs),
+            with_=withs.values())
 
     @classmethod
     def _joins(cls):
@@ -68,10 +72,11 @@ class Abstract(ModelSQL):
         tables['line'] = line = Line.__table__()
         tables['line.sale'] = sale = Sale.__table__()
         tables['line.sale.company'] = company = Company.__table__()
-        currency_sale = Currency.currency_rate_sql()
-        tables['currency_sale'] = currency_sale
-        currency_company = Currency.currency_rate_sql()
-        tables['currency_company'] = currency_company
+        withs = {}
+        currency_sale = With(query=Currency.currency_rate_sql())
+        withs['currency_sale'] = currency_sale
+        currency_company = With(query=Currency.currency_rate_sql())
+        withs['currency_company'] = currency_company
 
         from_item = (line
             .join(sale, condition=line.sale == sale.id)
@@ -88,20 +93,21 @@ class Abstract(ModelSQL):
                 & ((currency_company.end_date == Null)
                     | (currency_company.end_date >= sale.sale_date))
                 ))
-        return from_item, tables
+        return from_item, tables, withs
 
     @classmethod
-    def _columns(cls, tables):
+    def _columns(cls, tables, withs):
         line = tables['line']
         sale = tables['line.sale']
-        currency_company = tables['currency_company']
-        currency_sale = tables['currency_sale']
+        currency_company = withs['currency_company']
+        currency_sale = withs['currency_sale']
 
+        quantity = Coalesce(line.actual_quantity, line.quantity)
         revenue = cls.revenue.sql_cast(
-            Sum(line.quantity * line.unit_price
+            Sum(quantity * line.unit_price
                 * currency_company.rate / currency_sale.rate))
         return [
-            cls._column_id(tables).as_('id'),
+            cls._column_id(tables, withs).as_('id'),
             Literal(0).as_('create_uid'),
             CurrentTimestamp().as_('create_date'),
             cls.write_uid.sql_cast(Literal(Null)).as_('write_uid'),
@@ -112,17 +118,17 @@ class Abstract(ModelSQL):
             ]
 
     @classmethod
-    def _column_id(cls, tables):
+    def _column_id(cls, tables, withs):
         line = tables['line']
         return Min(line.id)
 
     @classmethod
-    def _group_by(cls, tables):
+    def _group_by(cls, tables, withs):
         sale = tables['line.sale']
         return [sale.company]
 
     @classmethod
-    def _where(cls, tables):
+    def _where(cls, tables, withs):
         context = Transaction().context
         sale = tables['line.sale']
 
@@ -167,10 +173,13 @@ class Abstract(ModelSQL):
         name = name[:-len('_trend')]
         if pygal:
             chart = pygal.Line()
-            chart.add('', [getattr(ts, name) if ts else None
+            chart.add('', [getattr(ts, name) if ts else 0
                     for ts in self.time_series_all])
             return chart.render_sparktext()
 
+    def get_currency(self, name):
+        return self.company.currency.id
+
     def get_currency_digits(self, name):
         return self.company.currency.digits
 
@@ -185,12 +194,12 @@ class AbstractTimeseries(Abstract):
         cls._order = [('date', 'ASC')]
 
     @classmethod
-    def _columns(cls, tables):
-        return super(AbstractTimeseries, cls)._columns(tables) + [
-            cls._column_date(tables).as_('date')]
+    def _columns(cls, tables, withs):
+        return super(AbstractTimeseries, cls)._columns(tables, withs) + [
+            cls._column_date(tables, withs).as_('date')]
 
     @classmethod
-    def _column_date(cls, tables):
+    def _column_date(cls, tables, withs):
         context = Transaction().context
         sale = tables['line.sale']
         date = DateTrunc(context.get('period'), sale.sale_date)
@@ -198,9 +207,9 @@ class AbstractTimeseries(Abstract):
         return date
 
     @classmethod
-    def _group_by(cls, tables):
-        return super(AbstractTimeseries, cls)._group_by(tables) + [
-            cls._column_date(tables)]
+    def _group_by(cls, tables, withs):
+        return super(AbstractTimeseries, cls)._group_by(tables, withs) + [
+            cls._column_date(tables, withs)]
 
 
 class Context(ModelView):
@@ -265,18 +274,25 @@ class Context(ModelView):
 
 
 class CustomerMixin(object):
-    customer = fields.Many2One('party.party', "Customer")
+    __slots__ = ()
+    customer = fields.Many2One(
+        'party.party', "Customer",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
 
     @classmethod
-    def _columns(cls, tables):
+    def _columns(cls, tables, withs):
         sale = tables['line.sale']
-        return super(CustomerMixin, cls)._columns(tables) + [
+        return super(CustomerMixin, cls)._columns(tables, withs) + [
             sale.party.as_('customer')]
 
     @classmethod
-    def _group_by(cls, tables):
+    def _group_by(cls, tables, withs):
         sale = tables['line.sale']
-        return super(CustomerMixin, cls)._group_by(tables) + [sale.party]
+        return super(CustomerMixin, cls)._group_by(tables, withs) + [
+            sale.party]
 
     def get_rec_name(self, name):
         return self.customer.rec_name
@@ -290,7 +306,12 @@ class Customer(CustomerMixin, Abstract, ModelView):
         'sale.reporting.customer.time_series', 'customer', "Time Series")
 
     @classmethod
-    def _column_id(cls, tables):
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('customer', 'ASC'))
+
+    @classmethod
+    def _column_id(cls, tables, withs):
         sale = tables['line.sale']
         return sale.party
 
@@ -301,23 +322,30 @@ class CustomerTimeseries(CustomerMixin, AbstractTimeseries, ModelView):
 
 
 class ProductMixin(object):
-    product = fields.Many2One('product.product', "Product")
+    __slots__ = ()
+    product = fields.Many2One(
+        'product.product', "Product",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
 
     @classmethod
-    def _columns(cls, tables):
+    def _columns(cls, tables, withs):
         line = tables['line']
-        return super(ProductMixin, cls)._columns(tables) + [
+        return super(ProductMixin, cls)._columns(tables, withs) + [
             line.product.as_('product')]
 
     @classmethod
-    def _group_by(cls, tables):
+    def _group_by(cls, tables, withs):
         line = tables['line']
-        return super(ProductMixin, cls)._group_by(tables) + [line.product]
+        return super(ProductMixin, cls)._group_by(tables, withs) + [
+            line.product]
 
     @classmethod
-    def _where(cls, tables):
+    def _where(cls, tables, withs):
         line = tables['line']
-        where = super(ProductMixin, cls)._where(tables)
+        where = super(ProductMixin, cls)._where(tables, withs)
         where &= line.product != Null
         return where
 
@@ -333,7 +361,12 @@ class Product(ProductMixin, Abstract, ModelView):
         'sale.reporting.product.time_series', 'product', "Time Series")
 
     @classmethod
-    def _column_id(cls, tables):
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('product', 'ASC'))
+
+    @classmethod
+    def _column_id(cls, tables, withs):
         line = tables['line']
         return line.product
 
@@ -344,14 +377,20 @@ class ProductTimeseries(ProductMixin, AbstractTimeseries, ModelView):
 
 
 class CategoryMixin(object):
-    category = fields.Many2One('product.category', "Category")
+    __slots__ = ()
+    category = fields.Many2One(
+        'product.category', "Category",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
 
     @classmethod
     def _joins(cls):
         pool = Pool()
         Product = pool.get('product.product')
         TemplateCategory = pool.get('product.template-product.category.all')
-        from_item, tables = super(CategoryMixin, cls)._joins()
+        from_item, tables, withs = super(CategoryMixin, cls)._joins()
         if 'line.product' not in tables:
             product = Product.__table__()
             tables['line.product'] = product
@@ -365,34 +404,37 @@ class CategoryMixin(object):
             from_item = (from_item
                 .join(template_category,
                     condition=product.template == template_category.template))
-        return from_item, tables
+        return from_item, tables, withs
 
     @classmethod
-    def _columns(cls, tables):
+    def _columns(cls, tables, withs):
         template_category = tables['line.product.template_category']
-        return super(CategoryMixin, cls)._columns(tables) + [
+        return super(CategoryMixin, cls)._columns(tables, withs) + [
             template_category.category.as_('category')]
 
     @classmethod
-    def _column_id(cls, tables):
+    def _column_id(cls, tables, withs):
+        pool = Pool()
+        Category = pool.get('product.category')
+        category = Category.__table__()
         line = tables['line']
         template_category = tables['line.product.template_category']
-        # Pairing function from http://szudzik.com/ElegantPairing.pdf
-        return Min(Case(
-                (line.id < template_category.id,
-                    (template_category.id * template_category.id) + line.id),
-                else_=(line.id * line.id) + line.id + template_category.id))
+        # Get a stable number of category over time
+        # by using number one order bigger.
+        nb_category = category.select(
+            Power(10, (Ceil(Log(Max(category.id))) + Literal(1))))
+        return Min(line.id * nb_category + template_category.id)
 
     @classmethod
-    def _group_by(cls, tables):
+    def _group_by(cls, tables, withs):
         template_category = tables['line.product.template_category']
-        return super(CategoryMixin, cls)._group_by(tables) + [
+        return super(CategoryMixin, cls)._group_by(tables, withs) + [
             template_category.category]
 
     @classmethod
-    def _where(cls, tables):
+    def _where(cls, tables, withs):
         template_category = tables['line.product.template_category']
-        where = super(CategoryMixin, cls)._where(tables)
+        where = super(CategoryMixin, cls)._where(tables, withs)
         where &= template_category.category != Null
         return where
 
@@ -408,7 +450,12 @@ class Category(CategoryMixin, Abstract, ModelView):
         'sale.reporting.category.time_series', 'category', "Time Series")
 
     @classmethod
-    def _column_id(cls, tables):
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('category', 'ASC'))
+
+    @classmethod
+    def _column_id(cls, tables, withs):
         template_category = tables['line.product.template_category']
         return template_category.category
 
@@ -433,6 +480,11 @@ class CategoryTree(ModelSQL, ModelView):
     currency_digits = fields.Function(
         fields.Integer("Currency Digits"), 'get_currency_digits')
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('name', 'ASC'))
+
     @classmethod
     def table_query(cls):
         pool = Pool()
@@ -446,6 +498,19 @@ class CategoryTree(ModelSQL, ModelView):
         categories = Category.browse(categories)
         return {c.id: c.name for c in categories}
 
+    @classmethod
+    def order_name(cls, tables):
+        pool = Pool()
+        Category = pool.get('product.category')
+        table, _ = tables[None]
+        if 'category' not in tables:
+            category = Category.__table__()
+            tables['category'] = {
+                None: (category, table.id == category.id),
+                }
+        return Category.name.convert_order(
+                'name', tables['category'], Category)
+
     def time_series_all(self):
         return []
 
@@ -467,12 +532,12 @@ class CategoryTree(ModelSQL, ModelView):
             sub_ids = list(sub_ids)
             where = reduce_ids(table.id, sub_ids)
             cursor.execute(*table.select(table.id, table.parent, where=where))
-            parents.update(cursor.fetchall())
+            parents.update(cursor)
 
             where = reduce_ids(reporting_category.id, sub_ids)
             cursor.execute(
                 *reporting_category.select(reporting_category.id, where=where))
-            reporting_categories.extend(r for r, in cursor.fetchall())
+            reporting_categories.extend(r for r, in cursor)
 
         result = {}
         reporting_categories = ReportingCategory.browse(reporting_categories)
@@ -513,37 +578,38 @@ class CategoryTree(ModelSQL, ModelView):
 
 
 class CountryMixin(object):
-
+    __slots__ = ()
     country = fields.Many2One('country.country', "Country")
 
     @classmethod
     def _joins(cls):
         pool = Pool()
         Address = pool.get('party.address')
-        from_item, tables = super(CountryMixin, cls)._joins()
+        from_item, tables, withs = super(CountryMixin, cls)._joins()
         if 'line.sale.shipment_address' not in tables:
             address = Address.__table__()
             tables['line.sale.shipment_address'] = address
             sale = tables['line.sale']
             from_item = (from_item
                 .join(address, condition=sale.shipment_address == address.id))
-        return from_item, tables
+        return from_item, tables, withs
 
     @classmethod
-    def _columns(cls, tables):
+    def _columns(cls, tables, withs):
         address = tables['line.sale.shipment_address']
-        return super(CountryMixin, cls)._columns(tables) + [
+        return super(CountryMixin, cls)._columns(tables, withs) + [
             address.country.as_('country')]
 
     @classmethod
-    def _group_by(cls, tables):
+    def _group_by(cls, tables, withs):
         address = tables['line.sale.shipment_address']
-        return super(CountryMixin, cls)._group_by(tables) + [address.country]
+        return super(CountryMixin, cls)._group_by(tables, withs) + [
+            address.country]
 
     @classmethod
-    def _where(cls, tables):
+    def _where(cls, tables, withs):
         address = tables['line.sale.shipment_address']
-        where = super(CountryMixin, cls)._where(tables)
+        where = super(CountryMixin, cls)._where(tables, withs)
         where &= address.country != Null
         return where
 
@@ -556,7 +622,12 @@ class Country(CountryMixin, Abstract):
         'sale.reporting.country.time_series', 'country', "Time Series")
 
     @classmethod
-    def _column_id(cls, tables):
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('country', 'ASC'))
+
+    @classmethod
+    def _column_id(cls, tables, withs):
         address = tables['line.sale.shipment_address']
         return address.country
 
@@ -570,25 +641,25 @@ class CountryTimeseries(CountryMixin, AbstractTimeseries, ModelView):
 
 
 class SubdivisionMixin(CountryMixin):
-
+    __slots__ = ()
     subdivision = fields.Many2One('country.subdivision', "Subdivision")
 
     @classmethod
-    def _columns(cls, tables):
+    def _columns(cls, tables, withs):
         address = tables['line.sale.shipment_address']
-        return super(SubdivisionMixin, cls)._columns(tables) + [
+        return super(SubdivisionMixin, cls)._columns(tables, withs) + [
             address.subdivision.as_('subdivision')]
 
     @classmethod
-    def _group_by(cls, tables):
+    def _group_by(cls, tables, withs):
         address = tables['line.sale.shipment_address']
-        return super(SubdivisionMixin, cls)._group_by(tables) + [
+        return super(SubdivisionMixin, cls)._group_by(tables, withs) + [
             address.subdivision]
 
     @classmethod
-    def _where(cls, tables):
+    def _where(cls, tables, withs):
         address = tables['line.sale.shipment_address']
-        where = super(SubdivisionMixin, cls)._where(tables)
+        where = super(SubdivisionMixin, cls)._where(tables, withs)
         where &= address.subdivision != Null
         return where
 
@@ -602,7 +673,12 @@ class Subdivision(SubdivisionMixin, Abstract):
         "Time Series")
 
     @classmethod
-    def _column_id(cls, tables):
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('subdivision', 'ASC'))
+
+    @classmethod
+    def _column_id(cls, tables, withs):
         address = tables['line.sale.shipment_address']
         return address.subdivision
 
@@ -667,21 +743,16 @@ class OpenRegion(Wizard):
 
     def transition_start(self):
         pool = Pool()
-        Region = pool.get('sale.reporting.region')
         Country = pool.get('sale.reporting.country')
         Subdivision = pool.get('sale.reporting.country.subdivision')
-        region = Region(Transaction().context['active_id'])
-        record = Region.union_unshard(region.id)
+        record = self.model.union_unshard(self.record.id)
         if isinstance(record, Country):
             return 'country'
         elif isinstance(record, Subdivision):
             return 'subdivision'
 
     def _do_action(self, action):
-        pool = Pool()
-        Region = pool.get('sale.reporting.region')
-        region = Region(Transaction().context['active_id'])
-        record = Region.union_unshard(region.id)
+        record = self.model.union_unshard(self.record.id)
         data = {
             'id': record.id,
             'ids': [record.id],
